package JSPservletPkg;

import java.util.*;
/**
 * Version 1.0.4. Mapper persistent cache support.<p>
 * Check if the user came back on routing page, which can
 * mean the target PageBox is not working.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 */
class BackCheck {
  /**
   * PageBox and timestamp wrapper.
   */
  class PBTS {
    /** PageBox URL. */
    String PageBoxUrl;
    /** Time when a requestor has been routed to the PageBox URL. */
    long time;
    /**
     * Constructor.
     * @param     PageBoxUrl PageBox URL.
     * @param     time time when a requestor has been routed to the PageBox URL.
     */
    PBTS(String PageBoxUrl, long time) {
      this.PageBoxUrl = PageBoxUrl;
      this.time = time;
    }
  }
  /**
   * Key: String - requestor URL.
   * Value: PBTS - PageBox and timestamp.
   */
  private HashMap lastRequestors = new HashMap();
  /** Embedded mapper. */
  private MapperIntf mapper = null;
  /** Handle on JSPhandler. */
  JSPhandler handler;
  /** Version 1.0.4. Mapper persistent cache support.<p>
   * Constructor creates Mapper.
   * @param     publisher publisher host name.
   */
  BackCheck(String publisher, JSPhandler handler) {
    this.handler = handler;
    mapper = new Mapper(publisher, handler);
  }
  /** Accessor. */
  final MapperIntf getMapper() {
    return mapper;
  }
  /**
   * Checks if a PageBox is up.
   * @param     requestor user host name.
   * @param     PageBoxUrl selected PageBox URL.
   * @return    true: can use it.
   */
  synchronized boolean check(String requestor, String PageBoxUrl) {
    handler.log.logprint(toString() + ".check(" + requestor + ", " + PageBoxUrl +
      ")");
    long cur = System.currentTimeMillis();
    PBTS pbts = null;
    if (((pbts = (PBTS)lastRequestors.get(requestor)) == null) ||
      !pbts.equals(PageBoxUrl)) {
      lastRequestors.put(requestor, new PBTS(PageBoxUrl, cur));
      return true;
    }
    Iterator i = lastRequestors.entrySet().iterator();
    while(i.hasNext()) {
      Map.Entry me = (Map.Entry)i.next();
      pbts = (PBTS)me.getValue();
      if (cur > (pbts.time + 300000 /* 5 mn */))
        i.remove();
    }
    if (!mapper.check(PageBoxUrl))
      return false;
    return true;
  }
}
