package JSPservletPkg;

/**
 * Cache interface for Mapper.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 */
interface CacheIntf {
  /** Returns a report string about cache content. */
  String report();
  /**
   * Register a new range and PageBox set in cache.<br>
   * Validated with mapper.testMisc.
   * @param address1 first requestor address in the range.
   * @param address2 last requestor address in the range.
   * @param code country code.
   */
  void register(long address1, long address2,
    String code);
  /**
   * Retrieve a country code from cache.<br>
   * Validated with mapper.testMisc.
   * @param address requestor address.
   */
  String retrieve(long address);
  /**
   * Remove a url mapping from cache.<br>
   * Validated with mapper.testMisc.
   * @param address requestor address - any address in the range.
   */
  void remove(long address);
  /**
   * Remove cache entries older than nbdays.
   * @param nbdays maximum life in cache (days).
   */
  void clear(int nbdays);
  /**
   * Periodically called.<br>
   * Intented to be used to save the cache.</p>
   */
  boolean save();
}