package JSPservletPkg;

import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.*;
import java.security.*;
/**
 * Class invoked by servlet container for a given context path.
 * It forwards the requests to a compiled jsp or servlet loaded by JSPloader.
 * <ul><li>Version 1.0.2: support of Tomcat 3.2.1 and SOAP for administration.</li>
 * <li>Version 1.0.3: mapper and routing support.</li>
 * <li>Version 1.0.4: routing support for forms, mappingProp support.</li></ul>
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 * @see JSPhandler
 * @see JSPhandler.ClassEntry
 * @see JSPloader
 */
public class JSPservlet extends HttpServlet {
  /** Synchronized HashMap
   * key: contextPath
   * value: JSPhandler
   */
  public static Map JSPhandlers = null;
  /** SimpleDateFormat used to build RFC conformant strings */
  private static final java.text.SimpleDateFormat sdf =
    new java.text.SimpleDateFormat("EEE, dd MMM yyyy hh:mm:ss");
  static {
    sdf.setTimeZone(TimeZone.getTimeZone("GMT"));
  }
  /**
   * <p>Changes<ul>
   * <li><p>Version 1.0.4<p>
   * In Resin 1.2.5, request.getContextPath() returns the exact path but
   * request.getAttribute("javax.servlet.include.context_path") returns the
   * lower case path. So we change the JSPhandlers to lower case.</li></ul></p>
   * <p>Returns the handling JSPhandler. Creates it the first time.
   * Synchronized method to protect the access to JSPhandlers.</p>
   * <p>In case of Resin, the JspFactory is normally not set at first call.
   * This factory is used in JSPs compiled with Resin.
   * I set it with Class.forName("com.caucho.jsp.QJspFactory").newInstance().
   * The code remains therefore portable across Java servers.<p>
   * @param     sc  ServletConfig used to initialize target classes.
   * @param     contextPath  application prefix.
   * @return    the requested JSPhandler.
   * @exception JSPloaderException
   *            JSPhandler constructor exception.
   * @see       JSPhandler
   */
  public static final synchronized JSPhandler getHandler(ServletConfig sc,
    String contextPath, ServletContext ctx, String URI, String URIRealPath)
    throws JSPloaderException {
    JSPhandler jh = null;
    if (JSPhandlers == null)
      JSPhandlers = Collections.synchronizedMap(new HashMap());
    else
      jh = (JSPhandler)JSPhandlers.get(contextPath.toLowerCase());
    if (jh != null)
      return jh;
    jh = new JSPhandler(sc, contextPath, ctx, URI, URIRealPath);
    JSPhandlers.put(contextPath.toLowerCase(), jh);
    jh.log.logprint("JSPservlet.getHandler(" + contextPath + ") " +
      JSPhandler.class.getClassLoader());
    setApplication();
    return jh;
  }
  /**
   * Servlet container specific initialization.
   * Current implementation usefull only in case of Resin to set up the default
   * factory.
   */
  public static final synchronized void setApplication() {
    if (javax.servlet.jsp.JspFactory.getDefaultFactory() != null)
      return;
    try {
      Class c = Class.forName("com.caucho.jsp.QJspFactory");
      if (c == null)
        return;
      javax.servlet.jsp.JspFactory.setDefaultFactory(
        (javax.servlet.jsp.JspFactory)c.newInstance());
    }
    catch(Exception e) {}
  }
  /**
   * Array of used HTTP codes.
   * Indexed by JSPhandler.EP_*.
   * @see       JSPhandler
   */
  static int[] aReason = {
    HttpServletResponse.SC_INTERNAL_SERVER_ERROR,
    HttpServletResponse.SC_NOT_FOUND,
    HttpServletResponse.SC_UNAUTHORIZED,
    HttpServletResponse.SC_UNAUTHORIZED};
  /**
   * If an error page is defined returns this page.
   * Otherwise returns an HTTP error.
   * The second method could be favoured if browsers and AS would behave
   * the same way.
   * As the page is specified in JSPservletPkg, I retrieve the page in the
   * JSPservletPkg war. Technically it would be possible to retrieve it in the
   * downloaded archive.
   * @param     handler JSPhandler.
   * @param     response HttpServletResponse.
   * @param     reason error code.
   * @see       JSPhandler
   */
  final void handleAbnormalCondition(JSPhandler handler,
    HttpServletResponse response, int reason) throws IOException {
    ServletContext sc = getServletContext();
    if ((handler == null) || (handler.aErrorPage[reason] == null) ||
      (sc == null)) {
      response.sendError(aReason[reason]);
      return;
    }
    handler.log.logprint(toString() + ".handleAbnormalCondition " +
      aReason[reason]);
    InputStream is = new ByteArrayInputStream(handler.aErrorPage[reason]);
    if (is == null) {
      handler.log.errorPrint(toString() + ".handleAbnormalCondition not found");
      response.sendError(aReason[reason]);
    }
    response.setContentType("text/html");
    OutputStream os = null;
    try {
      os = response.getOutputStream();
    }
    catch(IllegalStateException ise) {}
    if (os != null) {
      BufferedInputStream bis = new BufferedInputStream(is);
      BufferedOutputStream bos = new BufferedOutputStream(os);
      int i;
      while((i = bis.read()) != -1)
        bos.write(i);
      bos.flush();
      bis.close();
      return;
    }
    // if IllegalStateException, then getWriter should work...
    PrintWriter pw = null;
    try {
      pw = response.getWriter();
    }
    catch(IllegalStateException ise) {
      response.sendError(aReason[reason]);
      return;
    }
    BufferedReader br = new BufferedReader(new InputStreamReader(is));
    String ln = null;
    // In this case skip page content up to body.
    // Constraint: JSP or servlet should access resources either before
    // writing or after writing the body.
    // If your code fails to conform this requirement, revert to response mode.
    // Typically you get the problem for authorization and address it by
    // removing UnauthorizedPage parameter of web.xml.
    boolean isBodyFound = false;
    while ((ln = br.readLine()) != null) {
      if (!isBodyFound) {
        int idx;
        if ((idx = ln.indexOf("<body")) == -1)
          continue;
        isBodyFound = true;
        ln = ln.substring(idx);
        idx = ln.indexOf('>');
        pw.println(ln.substring(idx + 1));
        continue;
      }
      pw.println(ln);
    }
    br.close();
    pw.flush();
  }
  /* We ask the real path for an arbitrary name that must not exist. */
  final static String arbitraryForRealPath = "/P0a1g2e3B4o5x6";
  /** Version 1.0.4. Retrieves the URI (RequestURL).
   * Used in ClassPathLoader.getResource.<br>
   * Servlet container invoked service method.<br>
   * Handles JSPloaderException
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   * @exception ServletException from JSPservlet.
   */
  public void service(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
    JSPhandler handler = null;
    try {
      final String contextPath = getPathInfo(request, true);
      String URI = "" + HttpUtils.getRequestURL(request);
      String URIRealPath = getServletContext().getRealPath(arbitraryForRealPath);
      StringTokenizer st = new StringTokenizer(URI, "/");
      // the first token is http: without leading slash
      String URI2 = st.nextToken();
      String withoutSlash = contextPath;
      if (withoutSlash.startsWith("/"))
        withoutSlash = contextPath.substring(1);
      boolean found = false;
      while(st.hasMoreTokens()) {
        String tok = st.nextToken();
        URI2 += ("/" + tok);
        if (tok.compareToIgnoreCase(withoutSlash) == 0) {
          found = true;
          break;
        }
      }
      /* getRequestURL returns the initial URL (not the included one).
       * So with Resin and install on &lt;app-dir&gt; or with Tomcat with
       * install in Tomcat web.xml, we need to check. */
      if (!found) {
        String cp = request.getContextPath();
        if ((cp == null) || (cp.length() == 0)) {
          String pi = request.getPathInfo();
          if (pi == null)
            pi = request.getServletPath();
          withoutSlash = pi;
          if (withoutSlash.startsWith("/"))
            withoutSlash = withoutSlash.substring(1);
          int ix = withoutSlash.indexOf('/');
          if (ix != -1)
            cp = withoutSlash.substring(0, ix);
          else
            cp = "";  // shouldn't occur
        }
        int idx = URI.indexOf(cp);
        URI2 = URI.substring(0, idx + cp.length()) + contextPath;
      }
      String URIRealPath2 = URIRealPath.replace('\\', '/');
      URIRealPath2 = URIRealPath2.substring(0,
        URIRealPath2.lastIndexOf(arbitraryForRealPath));
      /*
      System.out.println(toString() + ".service getRequestURL:" + URI +
        " with contextPath:" + URI2 + " RealPath:" + URIRealPath +
        " without URI:" + URIRealPath2 + " contextPath:" + contextPath);
       */
      handler = getHandler(getServletConfig(), contextPath, getServletContext(),
        URI2, URIRealPath2);
      handler.log.logprint(toString() + ".service getRequestURL:" + URI +
        " with contextPath:" + URI2 + " RealPath:" + URIRealPath +
        " without URI:" + URIRealPath2 + " contextPath:" + contextPath);
      final JSPhandler acHandler = handler;
      final HttpServletRequest acRequest = request;
      final HttpServletResponse acResponse = response;
      AccessController.doPrivileged(new PrivilegedExceptionAction() {
        public Object run() throws Exception {
          myService(acRequest, acResponse, contextPath, acHandler);
          return null;
        }
      });
    }
    catch(PrivilegedActionException pae) {
      Exception e = pae.getException();
      if (e instanceof ClassNotFoundException)
        handleAbnormalCondition(handler, response, JSPhandler.EP_NOTFOUND);
      else
      if (e instanceof JSPloaderException) {
        JSPloaderException jle = (JSPloaderException)e;
        switch(jle.reason) {
          case JSPloaderException.JLE_UNDEF:
            handleAbnormalCondition(handler, response, JSPhandler.EP_NOTFOUND);
            break;
          case JSPloaderException.JLE_REVOKED:
            handleAbnormalCondition(handler, response, JSPhandler.EP_REVOKED);
            break;
          default:
            handleAbnormalCondition(handler, response, JSPhandler.EP_OTHER);
        }
      } else
          handleAbnormalCondition(handler, response, JSPhandler.EP_OTHER);
    }
    catch(JSPloaderException jle) {
      switch(jle.reason) {
        case JSPloaderException.JLE_UNDEF:
          handleAbnormalCondition(handler, response, JSPhandler.EP_NOTFOUND);
          break;
        case JSPloaderException.JLE_REVOKED:
          handleAbnormalCondition(handler, response, JSPhandler.EP_REVOKED);
          break;
        default:
          handleAbnormalCondition(handler, response, JSPhandler.EP_OTHER);
      }
    }
    catch(java.security.AccessControlException ace) {
      // ace.printStackTrace();
      handler.log.errorPrint(toString() + ".service Unauthorized call " + ace);
      handleAbnormalCondition(handler, response, JSPhandler.EP_UNAUTHORIZED);
    }
  }
  /**
   * Helper class.
   * Cleaner display in ServletUpdate if adds entries on its errorMsg ArrayList.
   */ 
  class IncludeWriter {
    /** response.getWriter if JSPupdate */
    PrintWriter pw;
    ArrayList errorMsgs = null;
    /**
     * Set pw if JSPupdate
     * @param     response HttpServletResponse.
     * @param     servletPath path.
     * @exception IOException only JSPupdate user stopped.
     */
    IncludeWriter(HttpServletResponse response, String servletPath)
      throws IOException{
      if (servletPath.equals("/JSPupdate"))
        pw = response.getWriter();
      else {
        pw = null;
        if (ServletUpdate.invoker != null)
          errorMsgs = ServletUpdate.invoker.errorMsgs;
      }
    }
    /**
     * Write on pw if JSPupdate
     * @param     msg message to display.
     */
    final void print(String msg) {
      if (pw != null)
        pw.println(ServletUpdate.msgHeader + msg + ServletUpdate.msgTrailer);
      else
      if (errorMsgs != null)
        errorMsgs.add(msg);
    }
  }
  /** Version 1.0.3.
   * <p>service method returning the response PrintWriter. Used in /ping and /info
   * handling</p>
   * @param     response HttpServletResponse.
   * @param     jh handling JSPhandler.
   */
  private final PrintWriter getWriter(HttpServletResponse response,
    JSPhandler jh) throws IOException {
    PrintWriter pw = null;
    try {
      pw = response.getWriter();
    }
    catch(IllegalStateException ise) {
      jh.log.errorPrint(toString() + ".getWriter illegal state");
      try {
        response.sendError(response.SC_FORBIDDEN);
      }
      catch(IOException ioe) {}
    }
    return pw;
  }
  /**
   * invokes the target servlet or handles an update on behalf of JSPupdate.
   * <p>myService is divided in:<ol>
   * <li>Checking for a JSPupdate invocation. In this case, it gets the jarName
   * from parameters and issues a JSPHandler update.</li>
   * <li>Checking for the extension. If there is no extension or the extension
   * is .class, then JSPhandler is invoked to get the servlet.</li>
   * <li>Otherwise JSPhandler is invoked to get the resource</li></ol>
   * <p>a JSP can be invoked through RequestDispatcher.include.
   * In this case request.getContextPath, request.getPathInfo point the including
   * servlet and not the included server, whose info are returned by
   * request.getAttribute("javax.servlet.include...").
   * <p>So I first look at request.getAttribute("javax.servlet.include...") and
   * then to request.getContextPath, request.getPathInfo.<p>
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   * @param     contextPath Context String retrieved from attribute (include) or
   *            from request.
   * @param     jh handling JSPhandler.
   * @exception ServletException internally raised.
   * @exception JSPloaderException from tool processing non recoverable.
   * @see       JSPhandler
   */
  private final void myService(HttpServletRequest request,
    HttpServletResponse response, String contextPath, JSPhandler jh)
    throws JSPloaderException, ClassNotFoundException, ServletException,
    IOException {
    String jarName = null;
    String servletPath = request.getServletPath();
    String includePath = JSPservlet.getIncludePrefix(request);
    /* jh.log.logprint(toString() + ".myService servletPath=" + servletPath +
      " " + "includePath=" + includePath); */
    // update handling (include case)
    if ((servletPath != null) &&
      (servletPath.equals(invokerMapping + "/ServletUpdate")
      /* Resin / app-dir case */
      || servletPath.equals(invokerMapping + "/JSPupdate")
      || servletPath.equals(includePath + "/JSPupdate")
      || servletPath.equals(includePath + "/ServletUpdate"))) {
      jarName = request.getParameter("jarName");
      String delete = request.getParameter("Delete");
      String remoteLocation = request.getParameter("remoteLocation");
      String publisher = request.getParameter("publisher");
      if (delete != null)
        jh.log.logprint(toString() + ".myService delete(" + jarName + ")");
      else
        jh.log.logprint(toString() + ".myService update(" + jarName + ", " +
          remoteLocation + ", " + publisher + ")");
      IncludeWriter iw = new IncludeWriter(response, servletPath);
      if (jarName == null)
        iw.print(jarName + " not set!");
      else {
        try {
          if (delete != null) {
            jh.delete(jarName);
            iw.print("JAR " + jarName + " deleted!");
          } else /* Version 1.0.3 */
          if ((publisher != null) && !jh.publisherUpdate(jarName, publisher))
            iw.print("JAR " + jarName + " published by another publisher (not " +
              publisher + ")");
          else {
            if (jh.update(jarName, remoteLocation))
              iw.print("JAR " + jarName + " updated!");
            else
              iw.print("JAR " + jarName + " loaded!");
          }
        }
        catch(JSPloaderException jle) {
          if (jle.reason == JSPloaderException.JLE_UNDEF)
            iw.print("JAR " + jarName +
              " not defined - set the remote location!");
          else
            iw.print("JAR " + jarName + " cannot be accessed!");
        }
      }
      return;
    }
    // resources and classes handling
	  String pathInfo = getPathInfo(request, false);
    // Version 1.0.3
    if (pathInfo.equals("/ping")) {
      PrintWriter pw = getWriter(response, jh);
      if (pw == null)
        return;
      pw.println("<html><body>" + jh.ID + " ping OK</html></body>");
      pw.flush();
      jh.log.logprint(toString() + ".myService ping(" + jh.ID + ")");
      return;
    }
    if (pathInfo.equals("/mapper")) {
      PrintWriter pw = getWriter(response, jh);
      if (pw == null)
        return;
      pw.println("<html><head><title>Mapper information</title></head><body>");
      pw.println("<form  action=mapper method=\"GET\">");
      pw.println("<table border=0 cellspacing=2 cellpadding=2 width=100%><tr>");
      pw.println("<td width=20%><strong>JAR Name :</strong></td>");
      pw.println("<td width=20%><select name=\"archive\">");
      Enumeration en = jh.remoteLocProp.propertyNames();
      while(en.hasMoreElements()) {
        String option = (String)en.nextElement();
        pw.println("<option value=\"" + option + "\">" + option);
      }
      pw.println(
        "</select></td><td width=20%></td><input type=\"Submit\" value=\"" +
        "show\" name=\"Show\"></tr></table></form>");
      jarName = request.getParameter("archive");
      if (jarName == null) {
        pw.println("archive not set</html></body>");
        pw.flush();
        return;
      }
      MapperIntf mapper = jh.getMapper(jarName);
      if (mapper == null) {
        pw.println("archive not found</html></body>");
        pw.flush();
        return;
      }
      pw.println("<h3>Mapper information</h3>" + mapper.getInfo());
      Properties p = jh.getRoutingPages(jarName);
      en = p.propertyNames();
      String msg = "<p><strong>Routing pages:</strong>";
      while(en.hasMoreElements()) {
        String key = (String)en.nextElement();
        msg += "<br>\n" + key + " " + p.getProperty(key);
      }
      pw.println(msg +
        "</p><hr>Alexis Grandemange (c) 2000-2001 GNU LGPL 2.1</html></body>");
      pw.flush();
      return;
    }
    String fullName = pathInfo;
    if (pathInfo.startsWith("/"))
      fullName = pathInfo.substring(1);
    String clName = jh.mappingProp.getProperty(fullName);
    if ((clName != null) && (jh.classPathLoader != null)) {
      Servlet s = jh.classPathLoader.cacheGet(clName);
      if (s == null) {
        Object o = null;
        try {
          o = jh.classPathLoader.loadClass(clName).newInstance();
        }
        catch(Exception e) {}
        if ((o != null) && (o instanceof Servlet)) {
          s = (Servlet)o;
          s.init(jh.servletConfig);
          jh.classPathLoader.cachePut(clName, s);
          jh.log.logprint(toString() + ".myService alias:" + fullName + " class:"
            + clName);
        }
      }
      if (s != null) {
        s.service(request, response);
        return;
      }
    }
    int idx = fullName.indexOf('/');
    if (idx == -1)
      throw new ServletException(toString() + ".service " + pathInfo +
        " without archive name");
    jarName = fullName.substring(0, idx);
    String classPath = fullName.substring(idx + 1);
    jh.log.logprint(toString() + ".myService contextPath " + contextPath +
      " pathInfo " + pathInfo);
    if ((idx = pathInfo.lastIndexOf('.')) != -1) {
      String suffix = pathInfo.substring(idx);
      /** Version 1.0.4. Support for servlets handling a suffix. */
      String className = jh.mappingProp.getProperty(suffix);
      if (className != null) {
        Servlet srv = jh.get(jarName, className);
        jh.log.logprint(toString() + ".service invoked servlet:" + srv + " for "
          + pathInfo);
        srv.service(request, response);
        return;
      }
      if (!pathInfo.endsWith(".class")) {
        /** Version 1.0.4. Support for ClassPathLoader resources. */
        long lastModified = -1;
        InputStream is = (jh.classPathLoader == null ? null :
          jh.classPathLoader.getResourceAsStream(fullName));
        if (is == null) {
          is = jh.getResourceAsStream(jarName, classPath);
          if (is == null) {
            response.sendError(response.SC_NOT_FOUND);
            return;
          }
          /* Version 1.0.3.
           * If no Session and if the resource is a routing resource
           * try href update.
           * Version 1.0.4. Session check in comments for Resin.
           */
          boolean inSession = true;
          // if (request.getSession(false) == null) {
            inSession = false;
            BackCheck backCheck = jh.getRoutingResource(jarName, classPath);
            if (backCheck != null) {
              hrefUpdate(jh, is, request, response, jarName, classPath, backCheck);
              return;
            }
          // }
          lastModified = jh.getLastModified(jarName, classPath);
        } else
          lastModified = jh.classPathLoader.getLastModified(fullName);
        jh.log.logprint(toString() + ".myService retrieved resource " + pathInfo);
        String strLastModified = sdf.format(new Date(lastModified));
        String strETag = "E" + lastModified;
        jh.log.logprint(toString() + ".myService resource(" + jarName + ", " +
          classPath + ") Last-Modified:" + strLastModified);
        String ifNoneMatch = request.getHeader("If-None-Match");
        if (ifNoneMatch != null) {
          jh.log.logprint(toString() + ".myService If-None-Match(" + jarName +
            ", " + classPath + ") ETag:" + strETag + "/" + ifNoneMatch);
          if (ifNoneMatch.equals(strETag)) {
            response.setDateHeader("Expires", System.currentTimeMillis() +
              jh.expiration * 1000);
            response.sendError(response.SC_NOT_MODIFIED);
            return;
          }
        }
        String ifModified = request.getHeader("If-Modified-Since");
        if (ifModified != null) {
          jh.log.logprint(toString() + ".myService If-Modified-Since(" + jarName
            + ", " + classPath + ") Last-Modified:" + strLastModified + "/" +
            ifModified);
          if (ifModified.equals(strLastModified)) {
            response.setDateHeader("Expires", System.currentTimeMillis() +
              jh.expiration * 1000);
            response.sendError(response.SC_NOT_MODIFIED);
            return;
          }
        }
        response.setHeader("Cache-control", "public");
        response.setDateHeader("Date", System.currentTimeMillis());
        response.setHeader("Last-Modified", strLastModified + " GMT");
        response.setHeader("ETag", "E" + lastModified);
        response.setDateHeader("Expires", System.currentTimeMillis() +
          jh.expiration * 1000);
        ServletContext sc = getServletContext();
        response.setContentType(sc.getMimeType(classPath));
        OutputStream os = null;
        try {
          os = response.getOutputStream();
        }
        catch(IllegalStateException ise) {
          jh.log.errorPrint(toString() + ".myService(" + jarName + ", " +
            classPath + ") getOutputStream failed");
        }
        if (os != null) {
          BufferedInputStream bis = new BufferedInputStream(is);
          BufferedOutputStream bos = new BufferedOutputStream(os);
          int j = 0;
          for(int i = 0;(i = bis.read()) != -1; ++ j)
            bos.write(i);
          bos.flush();
          bis.close();
          jh.log.logprint(toString() + ".myService(" + jarName + ", " +
            classPath + ") uses OutputStream written " + j);
          return;
        }
        PrintWriter pw = null;
        try {
          pw = response.getWriter();
        }
        catch(IllegalStateException ise) {
          jh.log.errorPrint(toString() + ".myService(" + jarName + ", " +
            classPath + ") illegal state");
          return;   // nothing we can do
        }
        jh.log.logprint(toString() + ".myService(" + jarName + ", " +
          classPath + ") uses Writer");
        BufferedReader br = new BufferedReader(new InputStreamReader(is));
        String ln = null;
        boolean isBodyFound = false, endBodyFound = false;
        while ((ln = br.readLine()) != null) {
          if (!isBodyFound) {
            if ((idx = indexOfIgnoreCase(ln, "<body")) == -1)
              continue;
            isBodyFound = true;
            ln = ln.substring(idx);
            idx = ln.indexOf('>');
            if (idx != -1) {
              pwprintln(jh, pw, ln.substring(idx + 1));
              endBodyFound = true;
            }
            continue;
          }
          if (!endBodyFound) {
            idx = ln.indexOf('>');
            if (idx != -1) {
              pwprintln(jh, pw, ln.substring(idx + 1));
              endBodyFound = true;
            }
            continue;
          }
          if ((idx = indexOfIgnoreCase(ln, "</body>")) != -1) {
            pwprintln(jh, pw, ln.substring(0, idx));
            break;
          }
          pwprintln(jh, pw, ln);
        }
        br.close();
        /* Flush only if we wrote.
         * I don't send response.SC_FORBIDDEN or response.SC_NO_CONTENT
         * because it would commit the response.
         */
        if (endBodyFound)
          pw.flush();
        return;
      }
      classPath = classPath.substring(0, idx);
    }
    Servlet srv = jh.get(jarName, classPath);
    jh.log.logprint(toString() + ".service invoked servlet:" + srv);
    srv.service(request, response);
  }
  /**
   * <ul><li>Version 1.0.3. href update of routing resources.</li>
   * <li>Version 1.0.4. action update of routing resources.</li></ul>
   * @param     jh handling JSPhandler.
   * @param     is input stream on routing resource.
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   * @param     jarName archive name.
   * @param     path path of routing resource.
   * @param     comment purpose of routing resource.
   */
  private void hrefUpdate(JSPhandler jh, InputStream is, HttpServletRequest request,
    HttpServletResponse response, String jarName, String path, BackCheck backCheck) {
    jh.log.logprint(toString() + ".hrefUpdate (" + jarName + ", " + path + ") ");
    PrintWriter pw = null;
    try {
      pw = response.getWriter();
    }
    catch(Exception e) {
      jh.log.errorPrint(toString() + ".hrefUpdate (" + jarName + ", " +
        path + ") " + e);
      try {
        response.sendError(response.SC_FORBIDDEN);
      }
      catch(IOException ioe) {}
      return;
    }
    String requestor = request.getRemoteHost();
    MapperIntf mapper = backCheck.getMapper();
    String PageBoxUrl = mapper.getDomain(requestor);
    if (PageBoxUrl != null)
      PageBoxUrl = mapper.getPageBox(PageBoxUrl, request.getRemoteAddr());
    if (PageBoxUrl == null)
      jh.log.logprint(toString() + ".hrefUpdate (" + jarName + ", " + path +
        ") found no PageBox " + mapper.getMessage() + " for " + requestor);
    else
      jh.log.logprint(toString() + ".hrefUpdate (" + jarName + ", " + path +
        ") selected " + PageBoxUrl + " for " + requestor);
    BufferedReader br = new BufferedReader(new InputStreamReader(is));
    String line = null;
    try {
      if ((PageBoxUrl == null) || (!backCheck.check(requestor, PageBoxUrl))) {
        while ((line = br.readLine()) != null)
          pw.println(line);
      } else
      while ((line = br.readLine()) != null) {
        String remain = line, toPrint = "";
        int idx = -1;
        while(true) {
          if ((idx = indexOfIgnoreCase(remain, "href=")) != -1) {
            idx += "HREF=".length();
            if ((remain.charAt(idx) == '\"') || (remain.charAt(idx) == '\''))
              ++ idx;
            if ((remain.charAt(idx) != '/') &&
              !remain.substring(idx).startsWith("http://"))
              toPrint += remain.substring(0, idx) + "http://" + PageBoxUrl + "/"
                + jarName + "/";
            else
              toPrint += remain.substring(0, idx);
            remain = remain.substring(idx);
          } else
          if ((idx = indexOfIgnoreCase(remain, "action=")) != -1) {
            idx += "ACTION=".length();
            if ((remain.charAt(idx) == '\"') || (remain.charAt(idx) == '\''))
              ++ idx;
            if ((remain.charAt(idx) != '/') &&
              !remain.substring(idx).startsWith("http://"))
              toPrint += remain.substring(0, idx) + "http://" + PageBoxUrl + "/"
                + jarName + "/";
            else
              toPrint += remain.substring(0, idx);
            remain = remain.substring(idx);
          } else
            break;
        }
        pw.println(toPrint + remain);
      }
    }
    catch(IOException ioe) {}
    pw.flush();
  }
  /** Version 1.0.2.
   * Service method to find a string both in upper and lower case.
   * The key is in in lower case.
   * @param     ln string where to find the string.
   * @param     key string to find.
   * @return    index.
   */
  private final int indexOfIgnoreCase(String ln, String key) {
    int idx = ln.indexOf(key);
    if (idx != -1)
      return idx;
    return ln.indexOf(key.toUpperCase());
  }
  /** Version 1.0.2.
   * Service method to write both on writer and log.
   * @param     jh handling JSPhandler.
   * @param     pw current print writer.
   */
  private final void pwprintln(JSPhandler jh, PrintWriter pw,
    String msg) {
    // jh.log.logprint(toString() + ".pwprintln " + msg);
    pw.println(msg);
  }
  /** Version 1.0.4.
   * Changes:<ul>
   * <li><p>Version 1.0.2</p>
   * <p>Empirical: depending on the application server, URI stuff goes either
   * to pathInfo or servlet path... The latter form fits better with 2.3 spec.
   * </p></li>
   * <li><p>Version 1.0.4</p>
   * <p>Resin support with WEB-INF on &lt;app-dir&gt;</p>
   * <p>Modif initially studied for Tomcat (JSPservlet defined in Tomcat
   * web.xml). With Tomcat you need a special InvokerServlet.</p></li></ul>
   * @param     request HttpServletRequest.
   * @param     bContextPath HttpServletRequest. If true, returns contextPath.
   *            if false, returns pathInfo without context path.
   * @return    a path containing the archive name and the class path separated
   *            by slashes.
   */
  final static String getPathInfo(HttpServletRequest request,
    boolean bContextPath) {
    String contextPath =
      (String)request.getAttribute("javax.servlet.include.context_path");
    // System.out.println("JSPservlet.getPathInfo context_path=" + contextPath);
    if (contextPath == null)
      contextPath = request.getContextPath();
    // System.out.println("JSPservlet.getPathInfo contextPath=" + contextPath);
    if (bContextPath && (contextPath != null) && (contextPath.length() > 0))
      return contextPath;
	  String pathInfo =
	    (String)request.getAttribute("javax.servlet.include.path_info");
    // System.out.println("JSPservlet.getPathInfo path_info=" + contextPath);
    if (pathInfo == null)
      pathInfo = request.getPathInfo();
    String servletPath = (String)request.getAttribute(
        "javax.servlet.include.servlet_path");
    if (servletPath == null)
      servletPath = request.getServletPath();
    /* System.out.println("JSPservlet.getPathInfo pathInfo=" + pathInfo +
      " servletPath=" + servletPath); */
    if (pathInfo == null)
      pathInfo = servletPath;
    else
    if (servletPath.startsWith(invokerMapping)) {
      if (!bContextPath)
        return pathInfo;
      else {
        String withoutSlash = servletPath;
        if (withoutSlash.startsWith("/"))
          withoutSlash = withoutSlash.substring(1);
        int idx = withoutSlash.indexOf('/');
        if (idx != -1)
          return withoutSlash.substring(idx);
        else
          return servletPath;
      }
    }
    String withoutSlash = pathInfo;
    if (withoutSlash.startsWith("/"))
      withoutSlash = withoutSlash.substring(1);
    int idx = withoutSlash.indexOf('/');
    if (idx != -1) {
      if (bContextPath)
        return "/" + withoutSlash.substring(0, idx);
      if ((contextPath == null) || (contextPath.length() == 0))
        return withoutSlash.substring(idx);
    }
    return pathInfo;
  }
  /* Version 1.0.4.
   * <p>Method used to cope with null contextPath.
   * It is used with Resin when WEB-INF is defined on &lt;app-dir&gt;.
   * It could also apply in case of Tomcat when define JSPservlet in Tomcat
   * web.xml. However you have to provide a special InvokerServlet.</p>
   * <p>Used in JSPservlet.service and ServletUpdate.</p>
   * <p>If contextPath is null, returns first part of the path, otherwise
   * returns "".</p>
   * <p>For Resin you must define JSPservlet mapping like this:
	 * &lt;servlet-name&gt;JSPservlet&lt;/servlet-name&gt;
	 * &lt;url-pattern&gt;/JSPservlet/*&lt;/url-pattern&gt;
   * instead of:
	 * &lt;servlet-name&gt;JSPservlet&lt;/servlet-name&gt;
	 * &lt;url-pattern&gt;/&lt;/url-pattern&gt;
   * @param     request HttpServletRequest.
   * @return    a prefix for RequestDispatcher.include.
   */
  final static String getIncludePrefix(HttpServletRequest request) {
    String contextPath =
      (String)request.getAttribute("javax.servlet.include.context_path");
    if (contextPath == null)
      contextPath = request.getContextPath();
    if ((contextPath != null) && (contextPath.length() > 0))
      return "";
	  String pathInfo =
	    (String)request.getAttribute("javax.servlet.include.path_info");
    if (pathInfo == null)
      pathInfo = request.getPathInfo();
    if (pathInfo == null)
      pathInfo = (String)request.getAttribute(
        "javax.servlet.include.servlet_path");
    if (pathInfo == null)
      pathInfo = request.getServletPath();
    if (pathInfo.startsWith("/"))
      pathInfo = pathInfo.substring(1);
    int idx = pathInfo.indexOf('/');
    if (idx == -1)
      return "";  // shouldn't happen - will fail later
    return "/" + pathInfo.substring(0, idx) + app_dir_ctx;
  }
  /**
   * String in:
	 * &lt;servlet-name&gt;JSPservlet&lt;/servlet-name&gt;
	 * &lt;url-pattern&gt;/JSPservlet/*&lt;/url-pattern&gt;
   */
  final static String app_dir_ctx = "/JSPservlet";
  /*
   * Must match invoker config. On Resin:
   * &lt;servlet-mapping url-pattern='/servlet/*' servlet-name='invoker'/&gt;
   */
  public final static String invokerMapping = "/servlet";
  /** service method for servlet using RequestDispatcher.
   * Its use is optional. It returns the jar name part of the path.
   * We recommend however its use as the servlet developer should not be aware
   * of its packaging.<p>
   * @param     request HttpServletRequest.
   * @return    the jar string.
   */
  public static final String getJAR(HttpServletRequest request) {
	  String pathInfo = getPathInfo(request, false);
    if (pathInfo == null)
      return "";
    if (pathInfo.startsWith("/"))
      pathInfo = pathInfo.substring(1);
    int idx = pathInfo.indexOf('/');
    if (idx == -1)
      return "";
    final String jarName = pathInfo.substring(0, idx);
    final String toReturn = "/" + jarName;
    final String classPath = pathInfo.substring(idx + 1);
    if (JSPhandlers == null)
      return "";
    final String contextPath = getPathInfo(request, true);
    final JSPhandler jh = (JSPhandler)JSPhandlers.get(contextPath.toLowerCase());
    if (jh == null)
      return "";
    jh.log.logprint("JSPservlet.getJAR contextPath " + contextPath +
      " jar " + jarName + " class " + classPath + " returns " + toReturn);
    Servlet srv = null;
    try {
      srv = jh.get(jarName, classPath);
    }
    catch(Exception e) {}
    if (srv == null)
      return "";
    return toReturn;
  }
  public String getServletInfo() {
    return "JSPservletPkg.JSPservlet Information";
  }
} 