package JSPservletPkg;
import java.util.*;
import java.net.*;
import java.io.*;
/**
 * Maps the user hostname.
 * Version 1.0.4.<ul>
 * <li>Global cache.</li>
 * <li>Cache to country code.</li>
 * <li>Persistent cache.</li>
 * <li>whois handled in a thread.</li></ul>
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 * @see Cache
 * @see whoisHandler
 */
class Mapper implements MapperIntf {
  /** Europa, Middle East and Africa region. Country code table. */
  static final Region EMEA = new Region("EMEA", new String[] {"AD" /* Andorra */,
    "AE" /* United Arab Emirates */, "AL" /* Albania */, "AO" /* Angola */,
    "AT" /* Austria */, "BA" /* Bosnia and Herzegovina */, "BE" /* Belgium */,
    "BF" /* Burkina Faso */, "BG" /* Bulgaria */, "BH" /* Bahrain */,
    "BI" /* Burundi */, "BJ" /* Benin */,
    "BW" /* Botswana */, "BY" /* Belarus */, "CF" /* Central African Republic */,
    "CG" /* Congo */, "CH" /* Switzerland */, "CI" /* Ivory Coast */,
    "CM" /* Cameroon */, "CS" /* Czechoslovakia */, "CV" /* Cape Verde */,
    "CY" /* Cyprus */, "CZ" /* Czech Republic */, "DE" /* Germany */,
    "DJ" /* Djibouti */, "DK" /* Denmark */, "DZ" /* Algeria */, "EE" /* Estonia */,
    "EG" /* Egypt */, "EH" /* Western Sahara */, "ER" /* Eritrea */, "ES" /* Spain */,
    "ET" /* Ethiopia */, "FI" /* Finland */, "FR" /* France */, "FX" /* France */,
    "GA" /* Gabon */, "GB" /* Great Britain */, "GH" /* Ghana */,
    "GI" /* Gibraltar */, "GM" /* Gambia */, "GN" /* Guinea */,
    "GQ" /* Equatorial Guinea */, "GR" /* Greece */, "GW" /* Guinea-Bissau */,
    "HR" /* Croatia */, "HU" /* Hungary */, "IE" /* Ireland */, "IL" /* Israel */,
    "IS" /* Iceland */, "IT" /* Italy */, "JO" /* Jordan */, "KE" /* Kenya */,
    "KW" /* Kuwait */, "LB" /* Lebanon */, "LI" /* Liechtenstein */,
    "LR" /* Liberia */, "LS" /* Lesotho */, "LT" /* Lithuania */,
    "LU" /* Luxembourg */, "LV" /* Latvia */, "LY" /* Libya */, "MA" /* Morocco */,
    "MC" /* Monaco */, "MD" /* Moldova */, "MG" /* Madagascar */,
    "MK" /* Macedonia */, "ML" /* Mali */, "MR" /* Mauritania */,
    "MS" /* Montserrat */, "MT" /* Malta */, "MW" /* Malawi */, "MZ" /* Mozambique */,
    "NA" /* Namibia */, "NE" /* Niger */, "NG" /* Nigeria */, "NL" /* Netherlands */,
    "NO" /* Norway */, "OM" /* Oman */, "PL" /* Poland */, "PT" /* Portugal */,
    "QA" /* Qatar */, "RE" /* Reunion */, "RO" /* Romania */, "RW" /* Rwanda */,
    "SA" /* Saudi Arabia */, "MU" /* Mauritius */, "SD" /* Sudan */, "SE" /* Sweden */,
    "SI" /* Slovenia */, "SK" /* Slovak Republic */, "SL" /* Sierra Leone */,
    "SM" /* San Marino */, "SN" /* Senegal */, "SO" /* Somalia */,
    "SY" /* Syria */, "SZ" /* Swaziland */, "TD" /* Chad */, "TG" /* Togo */,
    "TN" /* Tunisia */, "TR" /* Turkey */, "TZ" /* Tanzania */, "UA" /* Ukraine */,
    "UG" /* Uganda */, "UK" /* United Kingdom */, "VA" /* Vatican */,
    "YT" /* Mayotte */, "YU" /* Yugoslavia */, "ZA" /* South Africa */,
    "ZM" /* Zambia */, "ZR" /* Zaire */, "ZW" /* Zimbabwe */, "KM" /* Comoros */,
    "FO" /* Faroe Islands */, "SJ" /* Svalbard and Jan Mayen Islands */,
    "ST" /* Sao Tome and Principe */});

  /** Asia region. Country code table. */
  static final Region Asia = new Region("Asia", new String[] { "AF" /* Afghanistan */,
    "AM" /* Armenia */, "AZ" /* Azerbaijan */, "BD" /* Bangladesh */,
    "BT" /* Bhutan */, "BN" /* Brunei Darussalam */, "CN" /* China */,
    "GE" /* Georgia */, "HK" /* Hong Kong */,
    "ID" /* Indonesia */, "IN" /* India */, "IQ" /* Iraq */, "IR" /* Iran */,
    "JP" /* Japan */, "KG" /* Kyrgyzstan */, "KH" /* Cambodia */,
    "KP" /* North Korea */, "KR" /* South Korea */, "KZ" /* Kazakhstan */,
    "LK" /* Sri Lanka */, "MM" /* Myanmar */, "MN" /* Mongolia */, "MO" /* Macau */,
    "MY" /* Malaysia */, "NP" /* Nepal */, "NT" /* Neutral Zone */,
    "PK" /* Pakistan */, "RU" /* Russian Federation */, "SG" /* Singapore */,
    "SR" /* Suriname */, "SU" /* USSR */, "TH" /* Thailand */, "TJ" /* Tajikistan */,
    "TM" /* Turkmenistan */, "TW" /* Taiwan */, "UZ" /* Uzbekistan */,
    "VN" /* Viet Nam */, "YE" /* Yemen */, "LA" /* Laos */});

  /** Latin America region. Country code table. */
  static final Region LatinAmerica = new Region("Latin", new String[] {
    "AG" /* Antigua and Barbuda */, "AI" /* Anguilla */, "BZ" /* Belize */,
    "AN" /* Netherlands Antilles */, "AR" /* Argentina */, "BB" /* Barbados */,
    "BM" /* Bermuda */, "BO" /* Bolivia */, "BR" /* Brazil */, "BS" /* Bahamas */,
    "CL" /* Chile */, "CO" /* Colombia */, "CR" /* Costa Rica */, "CU" /* Cuba */,
    "DO" /* Dominican Republic */, "EC" /* Ecuador */, "GF" /* French Guiana */,
    "GD" /* Grenada */, "GP" /* Guadeloupe */, "GT" /* Guatemala */, "GY" /* Guyana */,
    "HN" /* Honduras */, "HT" /* Haiti */, "JM" /* Jamaica */,
    "KY" /* Cayman Islands */, "MQ" /* Martinique */, "MX" /* Mexico */,
    "NI" /* Nicaragua */, "FK" /* Malvinas */, "PA" /* Panama */, "PE" /* Peru */,
    "PR" /* Puerto Rico */, "PY" /* Paraguay */, "SV" /* El Salvador */,
    "UY" /* Uruguay */, "VC" /* Saint Vincent */, "VE" /* Venezuela */,
    "VG" /* Virgin Islands (British) */, "VI" /* Virgin Islands (U.S.) */,
    "TT" /* Trinidad and Tobago */, "CC" /* Cocos Islands */, "DM" /* Dominica */,
    "KN" /* Saint Kitts and Nevis */, "LC" /* Saint Lucia */,
    "TC" /* Turks and Caicos Islands */, "AW" /* Aruba */});

  /** Pacific Ocean region. Country code table. */
  static final Region Ocean = new Region("Ocean", new String[] {"AS"
    /* American Samoa */, "AQ" /* Antarctica */, "AU" /* Australia */, "FJ" /* Fiji */,
    "FM" /* Micronesia */, "GU" /* Guam */, "IO" /* British Indian Ocean Territory */,
    "MH" /* Marshall Islands */, "MP" /* Northern Mariana Islands */,
    "MV" /* Maldives */, "NC" /* New Caledonia */, "NZ" /* New Zealand */,
    "PF" /* French Polynesia */, "PG" /* Papua New Guinea */, "PH" /* Philippines */,
    "PN" /* Pitcairn */, "Sb" /* Solomon Islands */, "SC" /* Seychelles */,
    "TO" /* Tonga */, "TP" /* East Timor */, "CK" /* Cook Islands */,
    "TK" /* Tokelau */, "GS" /* S. Georgia and S. Sandwich Isls. */,
    "NR" /* Nauru */, "PW" /* Palau */, "TV" /* Tuvalu */, "CX" /* Christmas Island */,
    "VU" /* Vanuatu */, "WF" /* Wallis and Futuna */, "WS" /* Samoa */,
    "HM" /* Heard and McDonald Islands */, "KI" /* Kiribati */,
    "NF" /* Norfolk Island */, "NU" /* Niue */});

  /** We put there US, Canada and small isolated states. Country code table. */
  static final Region Misc = new Region("Misc", new String[] {
    "AC" /* Ascension island */, "BV" /* Bouvet Island */, "CA" /* Canada */,
    "GL" /* Greenland */, "PM" /* St. Pierre and Miquelon */, "SH" /* St. Helena */,
    "TF" /* French Southern Territories */, "UM" /* US Minor Outlying Islands */,
    "US" /* United States */, "MIL" /* US Military */, "NATO" /* Nato field */,
    "EDU" /* US Educational */, "GOV" /* US Government */, "INT" /* International */,
    "ARPA" /* Old style Arpanet */});

  /**
   * Europa, Middle East and Africa region. Country name table.
   * Indexes must match EMEA region.
   */
  static final Region EMEAnames = new Region("EMEA", new String[] {"Andorra",
    "United Arab Emirates", "Albania", "Angola", "Austria", "Bosnia and Herzegovina",
    "Belgium", "Burkina Faso", "Bulgaria", "Bahrain", "Burundi", "Benin",
    "Botswana", "Belarus", "Central African Republic", "Congo", "Switzerland",
    "Cote d\'Ivoire", "Cameroon", "Czechoslovakia", "Cape Verde", "Cyprus",
    "Czech Republic", "Germany", "Djibouti", "Denmark", "Algeria", "Estonia",
    "Egypt", "Western Sahara", "Eritrea", "Spain", "Ethiopia", "Finland", "France",
    "France", "Gabon", "Great Britain", "Ghana", "Gibraltar", "Gambia", "Equatorial Guinea",
    "Guinea-Bissau", "Guinea", "Greece", "Croatia", "Hungary", "Ireland", "Israel",
    "Iceland", "Italy", "Jordan", "Kenya", "Kuwait", "Lebanon", "Liechtenstein",
    "Liberia", "Lesotho", "Lithuania", "Luxembourg", "Latvia", "Libya", "Morocco",
    "Monaco", "Moldova", "Madagascar", "Macedonia", "Mali", "Mauritania",
    "Montserrat", "Malta", "Malawi", "Mozambique", "Namibia", "Niger", "Nigeria",
    "Netherlands", "Norway", "Oman", "Poland", "Portugal", "Qatar", "Reunion",
    "Romania", "Rwanda", "Saudi Arabia", "Mauritius", "Sudan", "Sweden", "Slovenia",
    "Slovak Republic", "Sierra Leone", "San Marino", "Senegal", "Somalia", "Syria",
    "Swaziland", "Chad", "Togo", "Tunisia", "Turkey", "Tanzania", "Ukraine",
    "Uganda", "United Kingdom", "Vatican", "Mayotte", "Yugoslavia", "South Africa",
    "Zambia", "Zaire", "Zimbabwe", "Comoros", "Faroe Islands",
    "Svalbard and Jan Mayen Islands", "Sao Tome and Principe"});

  /** Asia region. Country name table. Indexes must match Asia region. */
  static final Region AsiaNames = new Region("Asia", new String[] { "Afghanistan",
    "Armenia", "Azerbaijan", "Bangladesh", "Bhutan", "Brunei Darussalam", "China",
    "Georgia", "Hong Kong", "Indonesia", "India", "Iraq", "Iran", "Japan",
    "Kyrgyzstan", "Cambodia", "North Korea", "South Korea", "Kazakhstan",
    "Sri Lanka", "Myanmar", "Mongolia", "Macau", "Malaysia", "Nepal", "Neutral Zone",
    "Pakistan", "Russian Federation", "Singapore", "Suriname", "USSR", "Thailand",
    "Tajikistan", "Turkmenistan", "Taiwan", "Uzbekistan", "Viet Nam", "Yemen", "Laos"});

  /**
   * Latin America region. Country name table.
   * Indexes must match LatinAmerica region.
   */
  static final Region LatinAmericaNames = new Region("Latin", new String[] {
    "Antigua and Barbuda", "Anguilla", "Belize", "Netherlands Antilles", "Argentina",
    "Barbados", "Bermuda", "Bolivia", "Brazil", "Bahamas", "Chile", "Colombia",
    "Costa Rica", "Cuba", "Dominican Republic", "Ecuador", "French Guiana",
    "Grenada", "Guadeloupe", "Guatemala", "Guyana", "Honduras", "Haiti", "Jamaica",
    "Cayman Islands", "Martinique", "Mexico", "Nicaragua", "Falklands", "Malvinas",
    "Panama", "Peru", "Puerto Rico", "Paraguay", "El Salvador", "Uruguay",
    "Saint Vincent", "Venezuela", "Virgin Islands (British)", "Virgin Islands (U.S.)",
    "Trinidad and Tobago", "Cocos Islands", "Dominica", "Saint Kitts and Nevis",
    "Saint Lucia", "Turks and Caicos Islands", "Aruba"});

  /** Pacific ocean region. Country name table. Indexes must match Ocean region. */
  static final Region OceanNames = new Region("Ocean", new String[] { "American Samoa",
    "Antarctica", "Australia", "Fiji", "Micronesia", "Guam",
    "British Indian Ocean Territory", "Marshall Islands", "Northern Mariana Islands",
    "Maldives", "New Caledonia", "New Zealand", "French Polynesia", "Papua New Guinea",
    "Philippines", "Pitcairn", "Solomon Islands", "Seychelles", "Tonga", "East Timor",
    "Cook Islands", "Tokelau", "S. Georgia and S. Sandwich Isls.", "Nauru", "Palau",
    "Tuvalu", "Christmas Island", "Vanuatu", "Wallis and Futuna", "Samoa",
    "Heard and McDonald Islands", "Kiribati", "Norfolk Island", "Niue"});

  /**
   * We put there US, Canada and small isolated states. Country name table.
   * Indexes must match LatinAmerica region.
   */
  static final Region MiscNames = new Region("Misc", new String[] { "Ascension island",
    "Bouvet Island", "Canada", "Greenland", "St. Pierre and Miquelon", "St. Helena",
    "French Southern Territories", "US Minor Outlying Islands", "United States",
    "US Military", "NATO", "US Educational", "US Government", "International",
    "ARPA"});

  /** For test purpose. AGRANDEM is my host name. Replace by yours... */
  static final Region Test = new Region("Test", new String[] {"LOCALHOST",
    "AGRANDEM"});

  /** For consistency... No whois server will be invoked. */
  static final Region TestNames = new Region("Test", new String[] {"LOCALHOST",
    "Alexis Grandemange"});

  /** Generic name region. Generic code table. */
  static final Region Unknown = new Region("Unknown", new String[] { "COM"
    /* Commercial */, "NET" /* Network */, "ORG" /* Non-Profit Organization */,
    "AERO", "BIZ", "COOP", "INFO", "MUSEUM", "NAME", "PRO"});
  /** List of non-generic code regions. */
  static Region[] regions = new Region[] { EMEA, Asia, LatinAmerica, Ocean, Misc,
    Test };
  /** List of non-generic name regions. Indexes must match regions array */
  static Region[] regionNames = new Region[] { EMEAnames, AsiaNames,
    LatinAmericaNames, OceanNames, MiscNames, TestNames };
  /**
   * Key: domain.<br>
   * Value: String[]. Subscribers.
   */
  private Map subscribers = Collections.synchronizedMap(new HashMap());
  /**
   * Key: country.<br>
   * Value: String[]. Subscribers.
   */
  private Map subscribersByCountry = Collections.synchronizedMap(
    new HashMap());
  /**
   * Key: region.<br>
   * Value: String[]. Subscribers.
   */
  private Map subscribersByRegion = Collections.synchronizedMap(new HashMap());
  /** Publisher URL. */
  private String publisher;
  /** Handle on JSPhandler. */
  private JSPhandler handler;
  /** Returns info string. */
  public String getInfo() {
    String report = "<p><strong>Info " + publisher + " subscribers:</strong>";
    Iterator i = subscribers.entrySet().iterator();
    while(i.hasNext()) {
      Map.Entry me = (Map.Entry)i.next();
      report += "<br>\n" + me.getKey();
      String[] subs = (String[])me.getValue();
      for (int j = 0; j < subs.length; ++j)
        report += " " + subs[j];
    }
    report += "</p>\n<p><strong>Subscribers by country:</strong>";
    i = subscribersByCountry.entrySet().iterator();
    while(i.hasNext()) {
      Map.Entry me = (Map.Entry)i.next();
      report += "<br>\n" + me.getKey();
      String[] subs = (String[])me.getValue();
      for (int j = 0; j < subs.length; ++j)
        report += " " + subs[j];
    }
    report += "</p>\n<p><strong>Subscribers by region:</strong>";
    i = subscribersByRegion.entrySet().iterator();
    while(i.hasNext()) {
      Map.Entry me = (Map.Entry)i.next();
      report += "<br>\n" + me.getKey();
      String[] subs = (String[])me.getValue();
      for (int j = 0; j < subs.length; ++j)
        report += " " + subs[j];
    }
    report += "</p>\n<p><strong>Cache:</strong>";
    report += handler.cache.report();
    report += "</p>\n<p><strong>Checker:</strong>";
    i = checker.allSubscribers.entrySet().iterator();
    while(i.hasNext()) {
      Map.Entry me = (Map.Entry)i.next();
      boolean b = ((Boolean)me.getValue()).booleanValue();
      report += "<br>\n" + me.getKey() + (b ? " UP" : " DOWN");
    }
    return report;
  }
  /**
   * Returns a readable IP address in usual xxx.xxx.xxx.xxx format.
   * @param add numeric value.
   */
  static String formatAddress(long add) {
    int[] Bs = new int[4];
    long addr = add;
    for (int i = 3; i > 0; --i) {
      Bs[i] = (int)(addr % 256);
      addr /= 256;
    }
    Bs[0] = (int)addr;
    return "" + Bs[0] + "." + Bs[1] + "." + Bs[2] + "." + Bs[3];
  }
  /**
   * Constructor.
   * @param publisher publisher url.
   * @param handler handle on JSPhandler.
   */
  Mapper(String publisher, JSPhandler handler) {
    this.publisher = publisher;
    this.handler = handler;
    synchronized(getClass()) {
      if (checker == null)
        checker = new Checker();
      if (whoisHdlr == null)
        whoisHdlr = new whoisHandler(handler);
    }
    handler.log.logprint(toString() + ".Mapper(" + publisher + ") initialized");
  }
  /**
   * Update global subscriber state.<br>
   * At each loop the scanner thread increments the invocation number.<br>
   * Allows to do 1/12 of the checks at a call.
   * @param counterForPing invocation number.
   */
  public void checkAll(int counterForPing) {
    handler.log.logprint(toString() + ".checkAll(" + counterForPing + ")");
    checker.checkAll(counterForPing);
  }
  /**
   * Checks the state of a subscribing PageBox and updates its state.
   * @param subscriber subscribing PageBox.
   * @return boolean true if PageBox is up.
   */
  public boolean check(String subscriber) {
    handler.log.logprint(toString() + ".check(" + subscriber + ")");
    return checker.check(subscriber);
  }
  /** Checker object created in constructor. */
  static Checker checker = null;
  /** whoisHandler object created in constructor. */
  static whoisHandler whoisHdlr = null;
  /** No operation - wait. */
  private static final int NOOP = 0;
  /** Check availability - use of subscrib and rc. */
  private static final int CHECK = 1;
  /** Checker thread class. */
  class Checker extends Thread {
    /**
     * Map of subscribers.
     * Validated with mapper.testMisc.<dir>
     * <li>Key: subscriber.</li>
     * <li>Value: Boolean true: UP, false: DOWN.</li></dir>
     */
    private HashMap allSubscribers = new HashMap();
    /** Equals NOOP or CHECK. */
    private int action = NOOP;
    /** Used in case of check. */
    private String subscrib = null;
    /** Used in case of check. */
    private boolean rc = false;
    /** Used to check request completion. */
    private boolean completed = false;
    /** Protects again startup race condition. */
    private boolean isStarted = false;
    /** 1.0.4d: don't check itself. Local host name. */
    private String myHostName = null;
    /** Starts the thread. */
    Checker() {
      try {
        myHostName = InetAddress.getLocalHost().getHostName();
      }
      catch(Exception e) {}
      start();
    }
    /**
     * Require allSubscribers update.
     * @param counterForPing hint between to update only a part of allSubscribers.
     */
    private void checkAll(int counterForPing) {
      Object[] subs = allSubscribers.keySet().toArray();
      if ((subs == null) || (subs.length == 0))
        return;
      int cfp = counterForPing % subs.length;
      int nbToCheck = (subs.length / 12) + 1;
      for (int i = nbToCheck * cfp;
        ((i < (nbToCheck * (cfp + 1))) && (i < subs.length)); ++i) {
        check((String)subs[i]);
      }
    }
    /**
     * Checks if a subscriber is up.
     * @param subscriber subscriber url.
     * @return boolean true if subscriber is up.
     */
    private boolean check(String subscriber) {
      if ((myHostName != null) &&
        (subscriber.startsWith(myHostName + ":") ||
        subscriber.startsWith(myHostName + "/")))
        return true;
      synchronized(Mapper.class) {
        if (!isStarted)
        Thread.yield();
        if (!isStarted) {
          msg = "checker thread not started.";
          Mapper.this.handler.log.errorPrint(msg);
          return false;
        }
        synchronized(this) {
          subscrib = subscriber;
          completed = false;
          action = CHECK;
          notify();
        }
        for (int i = 0; i < 50; ++i) {
          try {
            Thread.sleep(20);
          }
          catch(InterruptedException ie) {}
          if (completed)
            return rc;
        }
        synchronized(this) {
          action = NOOP;
          notify();
        }
        msg = "subscriber don't answer";
        Mapper.this.handler.log.errorPrint(msg);
        return false;
      }
    }
    /** Main method of Thread. */
    public synchronized void run() {
      isStarted = true;
      while(true) {
        try {
          wait();
        }
        catch(InterruptedException ie) {}
        if (action == NOOP)
          continue;
        if (action == CHECK)
          rc = threadCheck(subscrib);
        completed = true;
        action = NOOP;
      }
    }
    /**
     * Returns available subscribers among a list.
     * @param subscribers array of subscribers.
     */
    private final synchronized String[] available(String[] subscribers) {
      ArrayList al = new ArrayList();
      for (int i = 0; i < subscribers.length; ++ i) {
        Boolean B = (Boolean)allSubscribers.get(subscribers[i]);
        if (B == null)
          continue;
        if (B.booleanValue())
          al.add(subscribers[i]);
      }
      int l = al.size();
      if (l == 0)
        return null;
      String[] rc = new String[l];
      Iterator i = al.iterator();
      for (int j = 0; i.hasNext(); ++ j)
        rc[j] = (String)i.next();
      return rc;
    }
    /**
     * Checks if a subscriber is up.
     * @param subscriber subscriber url.
     * @return boolean true if subscriber is up.
     */
    private boolean threadCheck(String subscriber) {
      boolean value = false;
      try {
        URL url = new URL("http://" + subscriber + "/ping");
        BufferedReader br = new BufferedReader(new InputStreamReader(
          url.openStream()));
        String line = null;
        while((line = br.readLine()) != null) {
          if (line.indexOf("ping OK") != -1) {
            br.close();
            value = true;
            break;
          }
        }
      }
      catch(Exception e) {
      }
      allSubscribers.put(subscriber, new Boolean(value));
      return value;
    }
  }
  /** Version 1.0.4. Cache persistence + cache per code.<p>
   * Initializes subscribers, subscribersByCountry, subscribersByRegion.<br>
   * Validated with mapper.testInit.
   */
  private boolean initialize() {
    handler.log.logprint(toString() + ".initialize");
    msg = null;
    if (!subscribers.isEmpty()) {
      Iterator i = whoisFailedSubscribers.iterator();
      while(i.hasNext()) {
        whoisFailedSubscriber wfs = (whoisFailedSubscriber)i.next();
        whoisHandler.CodeRange cr = whois(wfs.address);
        if (cr == null)
          continue;
        handler.cache.register(cr.add1, cr.add2, cr.code);
        register(cr.code, wfs.subscriber);
        i.remove();
      }
      return true;
    }
    HashMap subsAl = new HashMap();
    try {
      URL url = new URL("http://" + publisher + "?todo=subscribers");
      BufferedReader br = new BufferedReader(new InputStreamReader(
        url.openStream()));
      String line = null;
      while((line = br.readLine()) != null) {
        if (line.equals(
          "<html><head><title>Publisher</title></head><body><h2>Publisher</h2>"))
          continue;
        if (line.indexOf("Misconfigured repository") != -1) {
          msg = "publisher without repository";
          Mapper.this.handler.log.errorPrint(msg);
          return false;
        }
        if (line.equals("</body></html>"))
          break;
        // Support of app-dir (Resin)
        String su = line;
        if (su.endsWith(JSPservlet.invokerMapping))
          su = line + JSPservlet.app_dir_ctx;
        String domain = getDomain(su);
        if (domain == null) {
          msg = "Invalid subscriber: " + su;
          Mapper.this.handler.log.errorPrint(msg);
          continue;
        }
        String[] subs = (String[])subscribers.get(domain);
        if (subs == null) {
          subs = new String[] { su };
          subscribers.put(domain, subs);
        } else {
          String[] newSubs = new String[subs.length + 1];
          for (int i = 0; i < subs.length; ++ i)
            newSubs[i] = subs[i];
          newSubs[subs.length] = su;
          subscribers.put(domain, newSubs);
        }
        subsAl.put(su, domain);
      }
      br.close();
    }
    catch(MalformedURLException mue) {
      msg = "GET http://" + publisher + "&todo=subscribers " + mue;
      handler.log.errorPrint(msg);
      subscribers.clear();
      return false;
    }
    catch(IOException ioe) {
      msg = "GET http://" + publisher + "&todo=subscribers " + ioe;
      handler.log.errorPrint(msg);
      subscribers.clear();
      return false;
    }
    if (subscribers.isEmpty()) {
      msg = "no subscribers";
      handler.log.errorPrint(msg);
      return false;
    }
    Iterator it = subsAl.keySet().iterator();
    while(it.hasNext()) {
      String subscriber = (String)it.next();
      check(subscriber);
      String hostname = subscriber;
      int idx = subscriber.indexOf(':');
      if (idx != -1)
        hostname = subscriber.substring(0, idx);
      else {
        idx = subscriber.indexOf('/');
        if (idx != -1)
          hostname = subscriber.substring(0, idx);
      }
      idx = hostname.lastIndexOf('.');
      String code = hostname.substring(idx + 1);
      int numcode = -1;
      try {
        numcode = Integer.parseInt(code);
      }
      catch(NumberFormatException nfe) {}
      boolean bccTld = false;
      if (numcode == -1) {
        int i = 0;
        for (; i < Unknown.countries.length; ++i) {
          if (code.equalsIgnoreCase(Unknown.countries[i]))
            break;
        }
        if (i == Unknown.countries.length)
          bccTld = true;
      }
      if (!bccTld) {
        InetAddress ia = null;
        try {
          ia = InetAddress.getByName(hostname);
        }
        catch(UnknownHostException uhe) {}
        if (ia == null) {
          removeFromSubscribers(subscriber, (String)subsAl.get(subscriber));
          continue;
        }
        String address = ia.getHostAddress();
        code = handler.cache.retrieve(getLongAddress(address));
        if (code == null) {
          whoisHandler.CodeRange cr = whois(address);
          if (cr == null) {
            whoisFailedSubscribers.add(new whoisFailedSubscriber(subscriber,
              address));
            continue;
          }
          code = cr.code;
          handler.cache.register(cr.add1, cr.add2, cr.code);
        }
      }
      register(code, subscriber);
    }
    return true;
  }
  /**
   * whois servers can be down...
   * <p>whoisFailedSubscriber contains data needed to query again the
   * authority for a subscriber.</p>
   * <p>whoisFailedSubscriber objects are stored in whoisFailedSubscribers.</p>
   * Validated with mapper.testInit.
   */
  class whoisFailedSubscriber {
    /** Subscribing PageBox. */
    String subscriber;
    /** PageBox address. */
    String address;
    /**
     * Constructor.
     * @param subscriber subscribing PageBox.
     * @param address PageBox address.
     */
    whoisFailedSubscriber(String subscriber, String address) {
      this.subscriber = subscriber;
      this.address = address;
    }
  }
  /** Stores whoisFailedSubscriber elements. */
  ArrayList whoisFailedSubscribers = new ArrayList();
  /**
   * Registers a PageBox in the proper country and region.
   * @param code country code.
   * @param subscriber subscribing PageBox.
   */
  private final void register(String code, String subscriber) {
    String Ucode = code.toUpperCase();
    // subscribersByCountry handling
    String[] subs = (String[])subscribersByCountry.get(Ucode);
    if (subs == null) {
      subs = new String[] { subscriber };
      subscribersByCountry.put(Ucode, subs);
    } else {
      String[] newSubs = new String[subs.length + 1];
      for (int i = 0; i < subs.length; ++ i)
        newSubs[i] = subs[i];
      newSubs[subs.length] = subscriber;
      subscribersByCountry.put(Ucode, newSubs);
    }
    // subscribersByRegion handling
    boolean found = false;
    for (int i = 0; i < regions.length; ++i) {
      Region reg = regions[i];
      for (int j = 0; j < reg.countries.length; ++j) {
        if (Ucode.equals(reg.countries[j])) {
          subs = (String[])subscribersByRegion.get(reg.name);
          if (subs == null) {
            subs = new String[] { subscriber };
            subscribersByRegion.put(reg.name, subs);
          } else {
            String[] newSubs = new String[subs.length + 1];
            for (int k = 0; k < subs.length; ++ k)
              newSubs[k] = subs[k];
            newSubs[subs.length] = subscriber;
            subscribersByRegion.put(reg.name, newSubs);
          }
          found = true;
          break;
        }
      }
      if (found)
        break;
    } // subscribersByRegion handling
  }
  /**
   * Cleanup subscribers for invalid subscribers.
   * @param subscriber invalid subscriber.
   * @param domain invalid subscriber domain.
   */
  void removeFromSubscribers(String subscriber, String domain) {
    handler.log.logprint(toString() + ".removeFromSubscribers(" + subscriber +
      ", " + domain + ")");
    String[] subs = (String[])subscribers.get(domain);
    if (subs.length == 1) {
      subscribers.remove(domain);
      return;
    }
    String[] newSubs = new String[subs.length - 1];
    int j = 0;
    for (int i = 0; i < subs.length; ++ i) {
      if (!subscriber.equals(subs[i]))
        newSubs[j ++] = subs[i];
    }
    subscribers.put(domain, newSubs);
  }
  /** Error - diagnosis message. */
  private String msg = null;
  /** returns error - diagnosis message. */
  public String getMessage() {
    return msg;
  }
  /**
   * Returns the URL of the PageBox to use for a given URL.
   * <p>Intented use: JSPservlet uses ServletRequest.getRemoteHost to get
   * the name of the requestor and updates page HREF with the returned URL.</p>
   * <ol><li>Checks if there are PageBoxes in the same domain as requestor</li>
   * <li>If the TLD is not a country code, calls whois</li>
   * <li>Scan countries for PageBoxes</li></ol>
   * @param domain requestor domain.
   * @param address requestor IP address.
   */
  public String getPageBox(String domain, String address) {
    handler.log.logprint(toString() + ".getPageBox(" + domain + ", " + address +
      ")");
    if (!initialize())
      return null;
    if (domain.startsWith("IP"))
      return search(address);
    String[] PageBoxes = (String[])subscribers.get(domain);
    if (PageBoxes != null) {
      PageBoxes = checker.available(PageBoxes);
      if (PageBoxes != null)
        return loadBalance(PageBoxes);
    }
    String code = domain;
    int idx = domain.lastIndexOf('.');
    if (idx != -1)
      code = domain.substring(idx + 1);
    int i = 0;
    for (; i < Unknown.countries.length; ++i) {
      if (code.equalsIgnoreCase(Unknown.countries[i]))
        return search(address);
    }
    PageBoxes = scan(code);
    if (PageBoxes == null) {
      msg = "unknown country";
      handler.log.errorPrint(msg);
      return null;
    }
    PageBoxes = checker.available(PageBoxes);
    if (PageBoxes == null) {
      msg = "no PageBox available";
      handler.log.logprint(msg);
      return null;
    } else
      return loadBalance(PageBoxes);
  }
  /**
   * Given a country code returns:<ol>
   * <li>PageBoxes in the same country.</li>
   * <li>If not found, PageBoxes in the same region.</li></ol>
   * @param code country code.
   */
  private String[] scan(String code) {
    String Ucode = code.toUpperCase();
    for (int i = 0; i < regions.length; ++i) {
      Region reg = regions[i];
      for (int j = 0; j < reg.countries.length; ++j) {
        if (Ucode.equals(reg.countries[j])) {
          String[] PageBoxes = (String[])subscribersByCountry.get(reg.countries[j]);
          if (PageBoxes != null)
            return PageBoxes;
          PageBoxes = (String[])subscribersByRegion.get(reg.name);
          if (PageBoxes != null)
            return PageBoxes;
          break;
        }
      }
    }
    return null;
  }
  /** Number used for round-robbin load balancing. */
  int num = 0;
  /**
   * Round-robbin load balancing returning a url.
   * @param PageBoxes candidate urls.
   */
  private final String loadBalance(String[] PageBoxes) {
    String PageBox = PageBoxes[num % PageBoxes.length];
    if (num == Integer.MAX_VALUE)
      num = 0;
    else
      ++ num;
    return PageBox;
  }
  /**
   * Converts the IP address in a number more suitable for comparison.<br>
   * Validated with mapper.testRipe.
   * @param address IP address.
   */
  final static long getLongAddress(String address) {
    StringTokenizer st = new StringTokenizer(address, ".");
    int tokenNb = 0;
    if ((tokenNb = st.countTokens()) != 4)
      return -1;
    try {
      long rc = Long.parseLong(st.nextToken());
      if (rc == 0)
        return -1;
      while(st.hasMoreTokens())
        rc = rc * 256 + Integer.parseInt(st.nextToken());
      return rc;
    }
    catch(NumberFormatException nfe) {
      return -1;
    }
  }
  /** Version 1.0.4. Lock method.
   * Returns the country code and the alloted range of addresses for a address.
   * @param address requestor address.
   */
  private static synchronized whoisHandler.CodeRange whois(String address) {
    return whoisHdlr.whois(address);
  }
  /** Version 1.0.4.
   * Search best PageBox based on the requestor address.
   * @param address requestor address.
   */
  private String search(String address) {
    long add = getLongAddress(address);
    if (add == -1) {
      msg = "invalid address " + address;
      handler.log.errorPrint(msg);
      return null;
    }
    String code = handler.cache.retrieve(add);
    if (code == null) {
      whoisHandler.CodeRange cr = whois(address);
      if (cr == null)
        return null;
      handler.cache.register(cr.add1, cr.add2, cr.code);
      code = cr.code;
    }
    String[] PageBoxes = scan(code);
    if (PageBoxes == null) {
      msg = "found no PageBox for " + code + " (" + address + ")";
      handler.log.logprint(msg);
      return null;
    }
    PageBoxes = checker.available(PageBoxes);
    if (PageBoxes == null) {
      msg = "no PageBox available";
      handler.log.logprint(msg);
      return null;
    }
    return loadBalance(PageBoxes);
  }
  /**
   * Service method. Extract domain part from host name.<br>
   * Validated with mapper.testMisc.
   * @param     host host name.
   */
  public String getDomain(String host) {
    String hostWithoutPort = host;
    int idx = host.indexOf(":");
    if (idx != -1)
      hostWithoutPort = host.substring(0, idx);
    else {
      idx = host.indexOf('/');
      if (idx != -1)
        hostWithoutPort = host.substring(0, idx);
    }
    StringTokenizer st = new StringTokenizer(hostWithoutPort, ".");
    int tokenNb = st.countTokens();
    String[] tokens = new String[tokenNb];
    for (int i = 0; st.hasMoreTokens(); ++i)
      tokens[i] = st.nextToken();
    // if parse > 0, then it is an address. DNS has failed.
    int parsedInt = -1;
    try {
      parsedInt = Integer.parseInt(tokens[0]);
    }
    catch(NumberFormatException nfe) {}
    if (parsedInt > 0) {
      if (tokens.length != 4) {
        msg = "invalid domain name";
        handler.log.errorPrint(msg);
        return null;
      }
      return "IP" + tokens[0] + "." + tokens[1] + "." + tokens[2] + "." +
        tokens[3];
    }
    if (tokens.length <= 2)
      return hostWithoutPort;
    // Strip last qualifier
    String rc = tokens[1];
    for (int i = 2; i < tokens.length; ++i)
      rc += ("." + tokens[i]);
    return rc;
  }
}