package JSPservletPkg;

/** Version 1.0.4. save method.
 * User hostname mapper interface.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 */
interface MapperIntf {
  /**
   * Returns the URL of the PageBox to use for a given URL.
   * <p>Intented use: JSPservlet uses ServletRequest.getRemoteHost to get
   * the name of the requestor and updates page HREF with the returned URL.
   * The invoker should use only the last two or three last fields of the URL,
   * for instance proxad.net for nice-1-a7-14-82.dial.proxad.net.</p>
   * @param domain requestor domain.
   * @param address requestor IP address.
   */
  String getPageBox(String domain, String address);
  /**
   * Update global subscriber state.
   * At each loop the scanner thread increments the invocation number up to 12.
   * Then it reset the invocation number. Allows to do 1/12 of the checks at a
   * call.
   * @param counterForPing invocation number.
   */
  void checkAll(int counterForPing);
  /**
   * Checks the state of a subscribing PageBox and updates its state.
   * @param subscriber subscribing PageBox.
   * @return boolean true if PageBox is up.
   */
  boolean check(String subscriber);
  /** returns error - diagnosis message. */
  String getMessage();
  /** returns info string. */
  String getInfo();
  /**
   * Service method. Extract domain part from host name.
   * @param     host host name.
   */
  String getDomain(String host);
}
