package JSPservletPkg;

import javax.servlet.*;
import javax.servlet.http.*;
import java.io.*;
import java.util.*;
/**
 * Update servlet. Equivalent to JSPupdate.
 * <p>Changes:<ul>
 * <li>Version 1.0.1: delete method</li>
 * <li>Version 1.0.2: publish compatibility</li>
 * <li>Version 1.0.4: app-dir Resin support. Resin 1.2.5 support.</li></ul></p>
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 * @see JSPservlet
 */
public class ServletUpdate extends HttpServlet {
  /** Invoked by Servlet container. */
  public void init(ServletConfig config) throws ServletException {
    super.init(config);
  }
  /** HTML prefix of error and notification messages */
  final static String msgHeader = "<p><font color=#DB1260 size=4><i>";
  /** HTML suffix of error and notification messages */
  final static String msgTrailer = "</i></font>";
  /** message ArrayList. */
  ArrayList errorMsgs = new ArrayList();
  /**
   * Instance that required include.
   * @see JSPservlet
   */
  static ServletUpdate invoker = null;
  /**
   * GET request handling.
   * Steps:<ol><li>Retrieve jarName parameter.</li>
   * <li>Search handler. If found invokes it update method.</li>
   * <li>Otherwise, dispatches JSPservlet.</li></ol>
   * @param     request HttpServletRequest.
   * @param     response HttpServletResponse.
   */
  public void doGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
    errorMsgs.clear();
    String contextPath = JSPservlet.getPathInfo(request, true);
    boolean askForCreate = false;
    if ((JSPservlet.JSPhandlers == null) ||
      !JSPservlet.JSPhandlers.containsKey(contextPath.toLowerCase()))
      askForCreate = true;
    String submitVal = request.getParameter("Submit");
    String deleteVal = request.getParameter("Delete");
    if (request.getParameter("GC") != null)
      System.gc();
    else
    if ((submitVal != null) || (deleteVal != null)) {
      String jarName = request.getParameter("jarName");
      if (jarName == null || jarName.length() == 0)
        errorMsgs.add("No JAR name supplied!");
      else {
        String withoutExt = jarName;
        int index = jarName.lastIndexOf('.');
        if (index != -1)
      	  withoutExt = jarName.substring(0, index);
        if ((JSPservlet.JSPhandlers == null) ||
      	  (!JSPservlet.JSPhandlers.containsKey(contextPath.toLowerCase()))) {
          synchronized(ServletUpdate.class) {
            invoker = this;
      	    RequestDispatcher rd = getServletContext().getRequestDispatcher(
      	      JSPservlet.getIncludePrefix(request) + "/JSPservlet");
      	    rd.include(request, response);
            errorMsgs.add("handler created!");
          }
        } else {
	        JSPhandler jh = (JSPhandler)JSPservlet.JSPhandlers.get(
            contextPath.toLowerCase());
          try {
            if (deleteVal != null) {
              jh.delete(withoutExt);
              errorMsgs.add("JAR " + withoutExt + " deleted!");
            } else {  // Version 1.0.3
              String publisher = request.getParameter("publisher");
              if ((publisher != null) && !jh.publisherUpdate(jarName, publisher))
                errorMsgs.add("JAR " + jarName +
                  " published by another publisher (not " + publisher + ")");
              else {
                if (jh.update(withoutExt, request.getParameter("remoteLocation")))
                  errorMsgs.add("JAR " + withoutExt + " updated!");
                else
                  errorMsgs.add("JAR " + withoutExt + " loaded!");
              }
            }
          }
          catch(JSPloaderException jle) {
            if (jle.reason == JSPloaderException.JLE_UNDEF)
              errorMsgs.add("JAR " + withoutExt +
                " not defined - set the remote location!");
            else
              errorMsgs.add("JAR " + withoutExt + " cannot be accessed!");
          }
        }
      }
    }
    response.setContentType("text/html");
    PrintWriter out = new PrintWriter (response.getOutputStream());
    out.println("<html>");
    out.println("<head><title>ServletUpdate</title></head>");
    out.println("<body>");
    out.println("<h2><font color=#DB1260>Servlet update</font></h2><hr>");
    out.println("<form  action=ServletUpdate method=\"GET\">");
    out.println("<table border=0 cellspacing=2 cellpadding=2 width=80%><tr>");
    if (JSPservlet.JSPhandlers != null) {
      out.println(
"<td width=80%><font face=\"Helvetica\"><b>Loaded/</b>Defined</td></tr><tr>");
      out.println("<td width=15%><font face=\"Helvetica\"><b>Handler</b></td>");
      out.println("<td width=5%><font face=\"Helvetica\"><b>Trace</b></td>");
      out.println(
        "<td width=40%><font face=\"Helvetica\"><b>Download location</b></td>");
      out.println(
        "<td width=20%><font face=\"Helvetica\"><b>Archive</b></td>");
      out.println("<td width=10%><font face=\"Helvetica\">&nbsp;</td>");
      out.println("<td width=10%><font face=\"Helvetica\">&nbsp;</td></tr>");
      Set jhs = JSPservlet.JSPhandlers.keySet();
      Iterator jhit = jhs.iterator();
      while(jhit.hasNext()) {
        boolean isFirst = true;
        String path = (String)jhit.next();
        JSPhandler jh = (JSPhandler)JSPservlet.JSPhandlers.get(path);
        if (request.getParameterNames().hasMoreElements() &&
          ((submitVal == null) || submitVal.equals("Update")) &&
          ((deleteVal == null) || deleteVal.equals("Delete"))) {
          if (request.getParameter("trace" + path) != null)
            jh.log.setTrace(true);
          else
            jh.log.setTrace(false);
        }
        Set definedArchives = jh.remoteLocProp.keySet();
        Iterator defArchIt = definedArchives.iterator();
        while(defArchIt.hasNext()) {
          String archive = (String)defArchIt.next();
          if (request.getParameter("Upd" + path + ":" + archive) != null) {
            try {
              jh.update(archive, null);
              break;
            }
            catch(JSPloaderException jle) {
              errorMsgs.add("update " + path + "/" + archive + " " + jle);
              break;
            }
          } else
          if (request.getParameter("Del" + path + ":" + archive) != null) {
            jh.delete(archive);
            break;  // only one delete because ConcurrentModificationException
          }
        }
        Set ces = jh.classEntries.keySet();
        Iterator ceit = ces.iterator();
        ArrayList loadedArchives = new ArrayList();
        while(ceit.hasNext()) {
          String archive = (String)ceit.next();
          if (isFirst) {
            out.println("<tr><td width=15%><font face=\"Helvetica\"><b>" + path
              + "</b></td>");
            out.println(
              "<td width=5%><font face=\"Helvetica\"><input type=\"checkbox\" "
              + (jh.log.getTrace() ? "checked" : "") + " name=\"trace" + path +
              "\" value=on></td>");
            isFirst = false;
          } else {
    out.println("<tr><td width=15%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    out.println("<td width=5%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
          }
          out.println("<td width=40%><font face=\"Helvetica\"><b>" +
            (String)jh.remoteLocProp.getProperty(archive) + "</b></td>");
          out.println("<td width=20%><font face=\"Helvetica\"><b>" + archive
            + "</b></td>");
          out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Upd\" name=\"Upd"
            + path + ":" + archive + "\"></td>");
          out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Del\" name=\"Del"
            + path + ":" + archive + "\"></td></tr>");
          loadedArchives.add(archive);
        }
        Object [] loadedArchiveA = loadedArchives.toArray();
        defArchIt = definedArchives.iterator();
        while(defArchIt.hasNext()) {
          String archive = (String)defArchIt.next();
          int i = 0;
          for (; i < loadedArchiveA.length; ++i) {
            if (archive.equals((String)loadedArchiveA[i]))
              break;
          }
          if (i < loadedArchiveA.length)
            continue;
          if (isFirst) {
            out.println("<tr><td width=15%><font face=\"Helvetica\"><b>" + path
              + "</b></td>");
            out.println(
              "<td width=5%><font face=\"Helvetica\"><input type=\"checkbox\" "
              + (jh.log.getTrace() ? "checked" : "") + " name=\"trace" + path +
              "\" value=on></td>");
            isFirst = false;
          } else {
    out.println("<tr><td width=15%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    out.println("<td width=5%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
          }
          out.println("<td width=40%><font face=\"Helvetica\">" +
            (String)jh.remoteLocProp.getProperty(archive) + "</td>");
          out.println("<td width=20%><font face=\"Helvetica\">" + archive +
            "</td>");
          out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Load\" name=\"Upd"
            + path + ":" + archive + "\"></td>");
          out.println(
"<td width=10%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Del\" name=\"Del"
            + path + ":" + archive + "\"></td></tr>");
        }
      }
    }
    out.println(
"<tr><td width=100%><font face=\"Helvetica\"><b>Definition</b></td></tr><tr>");
    out.println("<td width=20%><font face=\"Helvetica\"><b>JAR Name :</b></td>");
    out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"text\" name=\"jarName\"></font></td>");
    out.println("<td width=60%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    out.println(
"</tr><tr><td width=20%><font face=\"Helvetica\"><b>Remote Location :</b></td>");
    out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"text\" name=\"remoteLocation\"></font></td>");
    out.println("<td width=60%><font face=\"Helvetica\"><b>&nbsp;</b></td>");
    if (askForCreate) {
      out.println(
"</tr><tr><td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Create\" name=\"Submit\"></td>");
      out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Remove\" name=\"Delete\"></td>");
    } else {
      out.println(
"</tr><tr><td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Update\" name=\"Submit\"></td>");
      out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"Delete\" name=\"Delete\"></td>");
      out.println(
"<td width=20%><font face=\"Helvetica\"><input type=\"Submit\" value=\"GC\" name=\"GC\"></td>");
    }
    out.println(
      "</tr></form></table><hr>Alexis Grandemange (c) 2000-2001 GNU LGPL 2.1");
    Object[] msgA = errorMsgs.toArray();
    for (int i = 0; i < msgA.length; ++i)
      out.println(msgHeader + msgA[i] + msgTrailer);
    out.println("</body></html>");
    out.close();
  }
  public String getServletInfo() {
    return "JSPservletPkg.ServletUpdate Information";
  }
} 