package JSPservletPkg;

import java.io.*;
import java.util.*;
import java.net.*;
import org.w3c.dom.*;
import org.apache.soap.util.xml.*;
import org.apache.soap.*;
import org.apache.soap.encoding.*;
import org.apache.soap.encoding.soapenc.*;
import org.apache.soap.rpc.*;
/**
 * Update SOAP class. Equivalent to ServletUpdate.
 * Changes:
 * Version 1.0.1: delete method
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author Alexis Grandemange based on Matthew J. Duftler (duftler@us.ibm.com) code
 * @version 1, 0, 1
 * @see JSPservlet
 */
class SoapClient {
  /**
   * Main method.
   * Exit codes:<ul><li>-1 Invalid parameter</li>
   * <li>-2 Soap exception</li>
   */
  public static void main(String[] args) throws Exception {
    if (args.length != 2) {
      System.err.println("Usage: SoapClient SOAP-router-URL contextPath");
      System.exit (-1);
    }
    URL url = new URL(args[0]);
    Call call = new Call();
    /*
    SOAPMappingRegistry smr = new SOAPMappingRegistry();
    call.setSOAPMappingRegistry(smr);
    */
    long begin = System.currentTimeMillis();
    for (int i = 0; i < 20; ++i) {
      call.setTargetObjectURI("urn:JSPservlet");
      init(call, url, args[1]);
      getList(call, url, "getLoaded");
      getList(call, url, "getDefined");
    }
    System.out.println(System.currentTimeMillis() - begin);
  }
  private static void init(Call call, URL url, String contextPath) {
    call.setMethodName("init");
    call.setEncodingStyleURI(Constants.NS_URI_SOAP_ENC);
    Vector params = new Vector();
    params.addElement(new Parameter("contextPath", String.class, contextPath,
      null));
    call.setParams(params);
    try {
      Response resp = call.invoke(url, "");
    }
    catch (SOAPException e) {
      System.err.println("init caught SOAPException (" + e.getFaultCode() +
        "): " + e.getMessage());
      e.printStackTrace();
      System.exit (-2);
    }
  }
  private static void getList(Call call, URL url, String method) {
    call.setMethodName(method);
    Vector params = new Vector();
    call.setParams(params);
    call.setEncodingStyleURI(Constants.NS_URI_LITERAL_XML);
    Response resp = null;
    try {
      resp = call.invoke(url, "");
    }
    catch (SOAPException e) {
      System.err.println(method + " caught SOAPException (" + e.getFaultCode() +
        "): " + e.getMessage());
      e.printStackTrace();
      System.exit (-2);
    }
    if (!resp.generatedFault()) {
      Parameter ret = resp.getReturnValue();
      Element loadedEl = (Element)ret.getValue();
      System.out.println(method + ": " + DOM2Writer.nodeToString(loadedEl));
    } else {
      Fault fault = resp.getFault();
      System.err.println(method + " generated fault: " + fault.getFaultCode() +
        " " + fault.getFaultString());
    }
  }
}