package JSPservletPkg;

import org.w3c.dom.*;
import org.apache.soap.util.xml.*;
import java.util.*;
/**
 * Update SOAP class. Equivalent to ServletUpdate.
 * <p>Changes:<ul>
 * <li>Version 1.0.4: Resin 1.2.5 support.</li></ul></p>
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0, 4
 * @see JSPservlet
 */
public class SoapUpdate {
  /** Target JSPhandler. */
  private JSPhandler handler = null;
  /** JSPservlet web archive path. */
  private String path;
  public SoapUpdate() {
    // System.out.println(toString() + ".SoapUpdate");
  }
  /**
   * Retrieves handler.
   * It cannot create it if it doesn't exist as it lacks a ServletConfig.
   * @param   contextPath JSPservlet web archive path.
   */
  public void init(String contextPath) {
    path = contextPath;
    if (contextPath == null) {
      // System.out.println(toString() + ".init null parameter");
      return;
    }
    if (JSPservlet.JSPhandlers == null) {
      /*
      System.out.println(toString() + ".init(" + contextPath +
        ") found null handlers " + getClass().getClassLoader());
       */
      return;
    }
    handler = (JSPhandler)JSPservlet.JSPhandlers.get(contextPath.toLowerCase());
    /*
    if (handler == null)
      System.out.println(toString() + ".init(" + contextPath +
        ") no handler found");
    System.out.println(toString() + ".init(" + contextPath + ") " + handler);
     */
  }
  /**
   * Set trace. Returns diagnosis message.
   * @param   OnOff ON (set trace on) or OFF (set trace off).
   */
  public String setTrace(String OnOff) {
    if (handler == null)
      return toString() + ".setTrace() no handler for " + path + ".";
    if (OnOff.equalsIgnoreCase("ON")) {
      handler.log.setTrace(true);
      return "trace set on.";
    }
    if (OnOff.equalsIgnoreCase("OFF")) {
      handler.log.setTrace(false);
      return "trace set off.";
    }
    return "Invalid parameter " + OnOff + ". Please enter ON or OFF.";
  }
  /**
   * Get trace. Returns ON or OFF or diagnosis message.
   */
  public String getTrace() {
    if (handler == null)
      return toString() + ".getTrace() no handler for " + path + ".";
    return (handler.log.getTrace() ? "ON" : "OFF");
  }
  /**
   * Update or creates an archive. Returns diagnosis message.
   * @param   jarName archive name.
   * @param   remoteLocation where to download the archive.
   */
  public String update(String jarName, String remoteLocation) {
    if (handler == null)
      return toString() + ".update() no handler for " + path + ".";
    try {
      return (handler.update(jarName, remoteLocation) ? jarName + " updated." :
        jarName + " added.");
    }
    catch(JSPloaderException jle) {
      return jle.getMessage();
    }
  }
  /**
   * Loads or reloads an archive. Returns diagnosis message.
   * @param   jarName archive name.
   */
  public String load(String jarName) {
    if (handler == null)
      return toString() + ".load() no handler for " + path + ".";
    try {
      handler.update(jarName, null);
    }
    catch(JSPloaderException jle) {
      return jle.getMessage();
    }
    return jarName + " loaded.";
  }
  /**
   * Removes an archive. Returns diagnosis message.
   * @param   jarName archive name.
   */
  public String delete(String jarName) {
    if (handler == null)
      return toString() + ".delete() no handler for " + path + ".";
    handler.delete(jarName);
    return jarName + " deleted.";
  }
  /**
   * Get list of loaded archives and remote locations.
   */
  public Element getLoaded() {
    XMLParserLiaison xpl = new XercesParserLiaison();
    Document doc = xpl.createDocument();
    Element archivesEl = doc.createElement("Archives");
    if (handler == null) {
      archivesEl.appendChild(doc.createTextNode("Null handler"));
      return archivesEl;
    }
    archivesEl.appendChild(doc.createTextNode("\n"));
    Set ces = handler.classEntries.entrySet();
    for (Iterator ceit = ces.iterator(); ceit.hasNext();) {
      Map.Entry me = (Map.Entry)ceit.next();
      String jarName = (String)me.getKey();
      JSPhandler.ClassEntry ce = (JSPhandler.ClassEntry)me.getValue();
      String remoteLocation = ce.jarURL;
      Element archiveEl = doc.createElement("Archive");
      Element nameEl = doc.createElement("Name");
      nameEl.appendChild(doc.createTextNode(jarName));
      Element rlEl = doc.createElement("Location");
      rlEl.appendChild(doc.createTextNode(remoteLocation));
      archiveEl.appendChild(doc.createTextNode("\n    "));
      archiveEl.appendChild(nameEl);
      archiveEl.appendChild(doc.createTextNode("\n    "));
      archiveEl.appendChild(rlEl);
      archiveEl.appendChild(doc.createTextNode("\n    "));
      archivesEl.appendChild(doc.createTextNode("  "));
      archivesEl.appendChild(archiveEl);
      archivesEl.appendChild(doc.createTextNode("\n"));
    }
    return archivesEl;
  }
  /**
   * Get list of defined archives and remote locations.
   */
  public Element getDefined() {
    XMLParserLiaison xpl = new XercesParserLiaison();
    Document doc = xpl.createDocument();
    Element archivesEl = doc.createElement("Archives");
    if (handler == null) {
      archivesEl.appendChild(doc.createTextNode("Null handler"));
      return archivesEl;
    }
    archivesEl.appendChild(doc.createTextNode("\n"));
    Set ces = handler.remoteLocProp.entrySet();
    for (Iterator ceit = ces.iterator(); ceit.hasNext();) {
      Map.Entry me = (Map.Entry)ceit.next();
      String jarName = (String)me.getKey();
      String remoteLocation = (String)me.getValue();
      if (handler.classEntries.containsKey(jarName))
        continue;
      Element archiveEl = doc.createElement("Archive");
      Element nameEl = doc.createElement("Name");
      nameEl.appendChild(doc.createTextNode(jarName));
      Element rlEl = doc.createElement("Location");
      rlEl.appendChild(doc.createTextNode(remoteLocation));
      archiveEl.appendChild(doc.createTextNode("\n    "));
      archiveEl.appendChild(nameEl);
      archiveEl.appendChild(doc.createTextNode("\n    "));
      archiveEl.appendChild(rlEl);
      archiveEl.appendChild(doc.createTextNode("\n    "));
      archivesEl.appendChild(doc.createTextNode("  "));
      archivesEl.appendChild(archiveEl);
      archivesEl.appendChild(doc.createTextNode("\n"));
    }
    return archivesEl;
  }
}
