package JSPservletPkg;

import java.util.*;
import java.net.*;
import java.io.*;
/**
 * Update client.
 * Usage: UpdateClient URL jarName [remoteLocation]
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0
 */
public class UpdateClient {
  /**
   * Main method.
   * Exit codes:<ul><li>-2 Invalid parameter</li>
   * <li>-1 Error</li><li> 1 Success</li></ul>
   * Steps:<ol><li>Build URL.</li>
   * <li>URL.openStream</li><li>Parse response</li></ol>
   */
  public static void main(String[] args) throws Exception {
    if (args.length < 2) {
      System.out.println("Usage: UpdateClient URL jarName [remoteLocation] (" +
        args.length + ")");
      System.exit(-2);
    }
    String url, jarName, remoteLocation = null;
    url = args[0];
    jarName = args[1];
    String fullURL = null;
    if (args.length == 3) {
      remoteLocation = args[2];
      fullURL = url + "/JSPupdate?jarName=" + jarName + "&remoteLocation="
        + URLEncoder.encode(remoteLocation) + "&Submit=JSPupdate";
    } else {
      fullURL = url + "/JSPupdate?jarName=" + jarName +
        "&remoteLocation=&Submit=JSPupdate";
    }
    System.out.println("INFO: UpdateClient URL:" + url + ",JAR name:" + jarName
      + ",remote location:" + remoteLocation + "\ncmd:" + fullURL);
    InputStream is = (new URL(fullURL)).openStream();
    if (is == null) {
      System.out.println("ERROR: Unable to connect to the Application Server");
      System.exit(-1);
    }
    int rc = -1;
    BufferedReader br = new BufferedReader(new InputStreamReader(is));
    while(true) {
      String line = br.readLine();
      if (line == null)
        break;
      if (line.indexOf("No JAR name supplied!") != -1) {
        System.out.println("ERROR: Invalid JAR name");
        break;
      }
      if ((line.indexOf("handler created!") != -1) ||
        (line.indexOf("JAR loaded!") != -1) ||
        (line.indexOf("JAR updated!") != -1)) {
        System.out.println("INFO: JAR updated");
        rc = 1;
        break;
      }
    }
    br.close();
    System.exit(rc);
  }
}
